
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <pthread.h>

class RefTarget;
class WeakRef;

class Hold {
public:
	Hold() : count_(0), waiting_(0)
	{ 
		pthread_mutex_init(&mutex_, NULL);
		pthread_cond_init(&cond_, NULL);
	}
	
	~Hold()
	{
		pthread_mutex_destroy(&mutex_);
		pthread_cond_destroy(&cond_);
	}

	void getHold() 
	{
		pthread_mutex_lock(&mutex_);
		++count_;
		pthread_mutex_unlock(&mutex_);
	}

	void releaseHold() 
	{
		pthread_mutex_lock(&mutex_);
		if (--count_ == 0 && waiting_ != 0)
			pthread_cond_broadcast(&cond_);
		pthread_mutex_unlock(&mutex_);
	}

	void wait()
	{
		pthread_mutex_lock(&mutex_);
		while(count_ != 0) {
			++waiting_;
			pthread_cond_wait(&cond_, &mutex_);
			--waiting_;
		}
		pthread_mutex_unlock(&mutex_);
	}

protected:
	pthread_mutex_t mutex_; // access to count_
	pthread_cond_t cond_; // signalled when count_ drops to 0
	int count_; // count of holds
	int waiting_; // count of threads waiting for release

private: // prevent the copy constructors and assignments
	Hold(const Hold&);
	void operator=(const Hold&);
};

class RefTarget {
	friend class WeakRef;
public:
	RefTarget(void *owner); // this pointer will be held in the weak references
	~RefTarget();

	WeakRef *newWeakRef(); // create a new weak reference

protected:
	void invalidate(); // called from the subclass destructor
	void freeWeakRef(WeakRef *wr); // notification from a WeakRef that it gets destroyed

	pthread_mutex_t mutex_; // controls all access
	Hold hold_;
	bool valid_; // gets reset on invalidation
	WeakRef *rlist_; // head of the list of references
	void *owner_; // value to pass to the weak refs

private: // prevent the default and copy constructors and assignments
	RefTarget();
	RefTarget(const RefTarget&);
	void operator=(const RefTarget&);
};

class WeakRef {
	friend class RefTarget;
public:
	~WeakRef();

	WeakRef *copy(); // create a copy of this reference

	void *grab(); // make the reference strong, return its current value
	void release(); // release the reference back to weak
	
protected:
	WeakRef(RefTarget *target); // called from copy() and from RefTarget

	bool invalidate1(); // called from RefTarget invalidation
	void invalidate2(); // called if invalidate1() returned true

	// a sort-of-double-linked list, access controlled by target's mutex
	WeakRef *next_;
	WeakRef **prevp_;

	pthread_mutex_t mutex_; // controls all access
	Hold hold_; // keeps track of holds from RefTarget
	Hold grabs_; // keep track of the grabs
	RefTarget *target_; // reset to 0 on invalidation

private: // prevent the default and copy constructors and assignments
	// (they may be added but the code is simpler without them)
	WeakRef();
	WeakRef(const WeakRef&);
	void operator=(const WeakRef&);
};

// ------------------ methods --------------------------

RefTarget::RefTarget(void *owner) : // this pointer will be held in the weak references
	valid_(true), rlist_(0), owner_(owner)
{
	pthread_mutex_init(&mutex_, NULL);
}

RefTarget::~RefTarget()
{
	if (valid_) {
		fprintf(stderr, "~RefTarget: not invalidated!\n");
		abort();
	}
	pthread_mutex_destroy(&mutex_);
}

WeakRef *RefTarget::newWeakRef() // create a new weak reference
{
	WeakRef *wr;
	wr = new WeakRef(this);
	pthread_mutex_lock(&mutex_);
	if (!valid_) {
		if (wr->invalidate1())
			wr->invalidate2();
	} else { // add to the list
		if (rlist_) {
			rlist_->prevp_ = &wr->next_;
		}
		wr->next_ = rlist_;
		wr->prevp_ = &rlist_;
		rlist_ = wr;
	}
	pthread_mutex_unlock(&mutex_);
	return wr;
}

void RefTarget::invalidate() // called from the subclass destructor
{
	pthread_mutex_lock(&mutex_);
	valid_ = false;
	while (rlist_) {
		WeakRef *wr = rlist_;
		rlist_ = rlist_->next_;
		if (rlist_)
			rlist_->prevp_ = &rlist_;
		wr->prevp_ = 0;
		wr->next_ = 0;
		bool code = wr->invalidate1();
		if (code) {
			pthread_mutex_unlock(&mutex_);
			wr->invalidate2();
			pthread_mutex_lock(&mutex_);
		}
	}
	pthread_mutex_unlock(&mutex_);
	hold_.wait();
}

void RefTarget::freeWeakRef(WeakRef *wr) // notification from a WeakRef that it gets destroyed
{
	// remove the ref from the list
	pthread_mutex_lock(&mutex_);
	if (wr->prevp_) {
		*wr->prevp_ = wr->next_;
		if (wr->next_)
			wr->next_->prevp_ = wr->prevp_;
		wr->prevp_ = 0;
		wr->next_ = 0;
	}
	pthread_mutex_unlock(&mutex_);
}

WeakRef::~WeakRef()
{
	pthread_mutex_lock(&mutex_);
	RefTarget *t = target_;
	target_ = 0;
	if (t)
		t->hold_.getHold();
	pthread_mutex_unlock(&mutex_);
	hold_.wait();
	if (t) {
		t->freeWeakRef(this);
		t->hold_.releaseHold();
	}
	pthread_mutex_destroy(&mutex_);
}

WeakRef *WeakRef::copy() // create a copy of this reference
{
	WeakRef *nwr;
	grab();
	pthread_mutex_lock(&mutex_);
	RefTarget *t = target_;
	pthread_mutex_unlock(&mutex_);
	if (t)
		nwr = t->newWeakRef();
	else
		nwr = new WeakRef(0);
	release();
	return nwr;
}

void *WeakRef::grab() // make the reference strong, return its current value
{
	grabs_.getHold();
	pthread_mutex_lock(&mutex_);
	void *owner = target_? target_->owner_ : 0;
	pthread_mutex_unlock(&mutex_);
	return owner;
}
void WeakRef::release() { // release the reference back to weak
	grabs_.releaseHold();
}

WeakRef::WeakRef(RefTarget *target) : // called from copy() and from RefTarget
	next_(0), prevp_(0), target_(target)
{
	pthread_mutex_init(&mutex_, NULL);
}

bool WeakRef::invalidate1() // called from RefTarget invalidation
{
	pthread_mutex_lock(&mutex_);
	RefTarget *t = target_;
	target_ = 0;
	if (t)
		hold_.getHold();
	pthread_mutex_unlock(&mutex_);
	return (t != 0);
}

void WeakRef::invalidate2() // called if invalidate1() returned true
{
	grabs_.wait();
	hold_.releaseHold();
}

// ------------------------- test -------------------------

pthread_mutex_t mutex; // for deletion

class Worker : public RefTarget {
public:
	Worker() : RefTarget(this)
	{ }

	~Worker()
	{
		RefTarget::invalidate();
		fprintf(stderr, "Worker: destroying %p\n", this);
	}

	static void *execute(void *arg)
	{
		Worker *w = (Worker *)arg;
		fprintf(stderr, "Worker: waiting %p\n", w);
		pthread_mutex_lock(&mutex);
		pthread_mutex_unlock(&mutex);
		delete w;
	}
};

void *
client1(void *arg)
{
	WeakRef *ref = (WeakRef *)arg;
	fprintf(stderr, "client1: weak ref %p\n", ref);
	for (int i = 0; i < 5; i++) {
		void *w = ref->grab();
		fprintf(stderr, "client1: got %p\n", w);
		sleep(1);
		fprintf(stderr, "client1: release\n");
		ref->release();
		sched_yield(); // Linux mutexes aren't honest...
	}
	delete ref;
}

void *
client2(void *arg)
{
	WeakRef *ref = (WeakRef *)arg;
	fprintf(stderr, "client2: weak ref %p\n", ref);
	fprintf(stderr, "client2: waiting\n");
	pthread_mutex_lock(&mutex);
	pthread_mutex_unlock(&mutex);
	delete ref;
	fprintf(stderr, "client2: deleted ref\n");
}

int main()
{
	pthread_t tw, tc1, tc2;

	pthread_mutex_init(&mutex, NULL);

	Worker *w = new Worker;
	WeakRef *ref = w->newWeakRef();

	pthread_mutex_lock(&mutex);

	pthread_create(&tw, NULL, Worker::execute, (void *)w);
	pthread_create(&tc1, NULL, client1, (void *)ref->copy());
	pthread_create(&tc2, NULL, client2, (void *)ref->copy());
	delete ref;

	sleep(3);
	pthread_mutex_unlock(&mutex);

	fprintf(stderr, "joining threads\n");
	pthread_join(tw, NULL);
	pthread_join(tc1, NULL);
	pthread_join(tc2, NULL);

	return 0;
}

/* 
The test output may vary depending on the scheduling order
but in general should be like:

Worker: waiting 0x503010
client1: weak ref 0x503300
client1: got 0x503010
client2: weak ref 0x503540
client2: waiting
client1: release
client1: got 0x503010
client1: release
client1: got 0x503010
joining threads
client2: deleted ref
client1: release
Worker: destroying 0x503010
client1: got (nil)
client1: release
client1: got (nil)
client1: release

*/
